#!/usr/bin/env python3
# -*- coding: utf-8 -*-

try:
    from importlib.resources import files
    from importlib.metadata import PackageNotFoundError
except (ModuleNotFoundError, ImportError) as e:
    from importlib_metadata import PackageNotFoundError
    from importlib_resources import files

from apppath import AppPath
from warg import clean_string, get_version, package_is_editable

__project__ = "Heimdallr"
__author__ = "Christian Heider Nielsen"
__version__ = "0.2.8"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

# __all__ = ["PROJECT_APP_PATH", "PROJECT_NAME", "PROJECT_VERSION", "get_version"]


PROJECT_NAME = clean_string(__project__)
PROJECT_VERSION = __version__
PROJECT_YEAR = 2019
PROJECT_AUTHOR = clean_string(__author__)
PROJECT_ORGANISATION = clean_string("Aivclab")
PROJECT_APP_PATH = AppPath(app_name=PROJECT_NAME, app_author=PROJECT_AUTHOR)

PACKAGE_DATA_PATH = files(PROJECT_NAME) / "data"

try:
    DEVELOP = package_is_editable(PROJECT_NAME)
except PackageNotFoundError as e:
    DEVELOP = True

__version__ = get_version(__version__, append_time=DEVELOP)
__version_info__ = tuple(int(segment) for segment in __version__.split("."))

if __name__ == "__main__":
    print(__project__, __version_info__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 19/03/2020
           """


import copy
import datetime
import json
import logging
import socket
from typing import Any

import dash
import flask
from dash import Dash
from dash.dash_table import DataTable
from dash.dependencies import Input, Output
from dash.html import Div
from draugr.writers import LogWriter, MockWriter, Writer
from flask import Response
from heimdallr.configuration.heimdallr_config import ALL_CONSTANTS
from heimdallr.configuration.heimdallr_settings import (
    HeimdallrSettings,
    SettingScopeEnum,
)
from heimdallr.server.board_layout import get_root_layout
from heimdallr.utilities.server import (
    get_calender_df,
    per_machine_per_device_pie_charts,
    to_overall_gpu_process_df,
)
from paho import mqtt
from paho.mqtt.client import Client
from pandas import DataFrame
from waitress import serve
from warg import NOD, default_datetime_repr
from warg import ensure_existence

from heimdallr import PROJECT_APP_PATH, PROJECT_NAME

__all__ = ["main"]

from heimdallr.utilities.server.du_utilities import to_overall_du_process_df
from heimdallr.utilities.server.teams_status import team_members_status

log = logging.getLogger("werkzeug")
log.setLevel(logging.ERROR)

# external JavaScript files
external_scripts = [
    "https://www.google-analytics.com/analytics.js",
    {"src": "https://cdn.polyfill.io/v2/polyfill.min.js"},
    {
        "src": "https://cdnjs.cloudflare.com/ajax/libs/lodash.js/4.17.10/lodash.core.js",
        "integrity": "sha256-Qqd/EfdABZUcAxjOkMi8eGEivtdTkh3b65xCZL4qAQA=",
        "crossorigin": "anonymous",
    },
]

# external CSS stylesheets
external_stylesheets = [
    "https://codepen.io/chriddyp/pen/bWLwgP.css",
    {
        "href": "https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/css/bootstrap.min.css",
        "rel": "stylesheet",
        "integrity": "sha384-MCw98/SFnGE8fJT3GXwEOngsV7Zt27NXFoaoApmYm81iuXoPkFOJwJ8ERdknLPMO",
        "crossorigin": "anonymous",
    },
]

GPU_STATS = NOD()
DU_STATS = NOD()
KEEP_ALIVE = NOD()

# CLIENT_ID = str(uuid.getnode())
HOSTNAME = socket.gethostname()
CLIENT_ID = HOSTNAME
MQTT_CLIENT = Client(
    client_id=CLIENT_ID,
    # clean_session=True
)
DASH_APP = Dash(
    __name__,
    external_scripts=external_scripts,
    external_stylesheets=external_stylesheets,
)

DEVELOPMENT = False
DASH_APP.layout = get_root_layout(DEVELOPMENT)
LOG_WRITER: Writer = MockWriter()


@DASH_APP.callback(
    Output(ALL_CONSTANTS.TIME_ID, "children"),
    [Input(ALL_CONSTANTS.TIME_INTERVAL_ID, "n_intervals")],
)
def update_time(n: int) -> str:
    """description"""
    global GPU_STATS
    global KEEP_ALIVE
    try:
        for k, v in KEEP_ALIVE.items():
            if v > ALL_CONSTANTS.TIMEOUT_MACHINES_SEC:
                if k in GPU_STATS.keys():
                    GPU_STATS.pop(k)
                    print(f"deleting {k} from stats")

        for key in list(KEEP_ALIVE.keys()):
            if key in GPU_STATS.keys():
                KEEP_ALIVE[key] = KEEP_ALIVE[key] + 1
            else:
                KEEP_ALIVE.pop(key)
                print(f"deleting {key} from keep alive")

    except Exception as e:
        print(e)

    return default_datetime_repr(datetime.datetime.now())


@DASH_APP.callback(
    Output(ALL_CONSTANTS.CALENDAR_ID, "children"),
    [Input(ALL_CONSTANTS.CALENDAR_INTERVAL_ID, "n_intervals")],
)
def update_calendar_live(n: int) -> DataTable:
    """description"""
    try:
        df = get_calender_df(
            HeimdallrSettings().google_calendar_id,
            HeimdallrSettings()._credentials_base_path,
            num_entries=ALL_CONSTANTS.TABLE_PAGE_SIZE,
        )

        return DataTable(
            id="calender-table-0",
            columns=[{"name": i, "id": i} for i in df.columns],
            data=df.to_dict("records"),
            page_size=ALL_CONSTANTS.TABLE_PAGE_SIZE,
            style_as_list_view=True,
            style_data_conditional=[
                {
                    "if": {"column_id": "start", "filter_query": "{start} > 3.9"},
                    "backgroundColor": "green",
                    "color": "white",
                },
            ],  # TODO: MAKE GRADIENT TO ORANGE FOR WHEN NEARING START, and GREEN WHEN IN PROGRESS
        )
    except Exception as e:
        return Div([f"Error: {e}"])


@DASH_APP.callback(
    Output(ALL_CONSTANTS.TEAMS_STATUS_ID, "children"),
    [Input(ALL_CONSTANTS.TEAMS_STATUS_INTERVAL_ID, "n_intervals")],
)
def update_teams_status_live(n: int) -> Div:
    """description"""

    return Div(team_members_status(None), className="row")


@DASH_APP.callback(
    Output(ALL_CONSTANTS.GPU_GRAPHS_ID, "children"),
    [Input(ALL_CONSTANTS.GPU_INTERVAL_ID, "n_intervals")],
)
def update_graph(n: int) -> Div:
    """description"""
    global GPU_STATS
    global KEEP_ALIVE
    compute_machines = []
    if GPU_STATS:
        compute_machines.extend(
            per_machine_per_device_pie_charts(copy.deepcopy(GPU_STATS), KEEP_ALIVE)
        )
    return Div(compute_machines)


@DASH_APP.callback(
    Output(ALL_CONSTANTS.GPU_TABLES_ID, "children"),
    [Input(ALL_CONSTANTS.GPU_INTERVAL_ID, "n_intervals")],
)
def update_table(n: int) -> Div:
    """description"""
    MQTT_CLIENT.loop()

    compute_machines = []

    if GPU_STATS:
        df = to_overall_gpu_process_df(copy.deepcopy(GPU_STATS))
    else:
        df = DataFrame(["No data"], columns=("data",))

    compute_machines.append(
        DataTable(
            id="gpu-table-0",
            columns=[{"name": i, "id": i} for i in df.columns],
            data=df.to_dict("records"),
            page_size=ALL_CONSTANTS.TABLE_PAGE_SIZE,
            # style_as_list_view=True,
            style_data_conditional=[
                {"if": {"row_index": "odd"}, "backgroundColor": "rgb(248, 248, 248)"}
            ],
            style_header={
                "backgroundColor": "rgb(230, 230, 230)",
                "fontWeight": "bold",
            },
        )
    )

    return Div(compute_machines)


@DASH_APP.callback(
    Output(ALL_CONSTANTS.DU_TABLES_ID, "children"),
    [Input(ALL_CONSTANTS.DU_INTERVAL_ID, "n_intervals")],
)
def update_table(n: int) -> Div:
    """description"""
    MQTT_CLIENT.loop()

    compute_machines = []

    if DU_STATS:
        df = to_overall_du_process_df(copy.deepcopy(DU_STATS))
    else:
        df = DataFrame(["No data"], columns=("data",))

    compute_machines.append(
        DataTable(
            id="du-table-0",
            columns=[{"name": i, "id": i} for i in df.columns],
            data=df.to_dict("records"),
            page_size=ALL_CONSTANTS.TABLE_PAGE_SIZE,
            # style_as_list_view=True,
            style_data_conditional=[
                {"if": {"row_index": "odd"}, "backgroundColor": "rgb(248, 248, 248)"}
            ],
            style_header={
                "backgroundColor": "rgb(230, 230, 230)",
                "fontWeight": "bold",
            },
        )
    )

    return Div(compute_machines)


@DASH_APP.callback(
    dash.dependencies.Output("menu_container", "style"),
    [dash.dependencies.Input("menu_toggle_button", "n_clicks")],
)
def menu_toggle(n_clicks: int) -> dict:
    """

    Args:
      n_clicks:

    Returns:

    """
    if n_clicks and n_clicks % 2 == 1:
        return {"display": "block"}
    return {"display": "none"}


@DASH_APP.server.route("/config", methods=["POST"])
def on_post_config() -> Response:
    """

    Returns:

    """
    if DEVELOPMENT:
        settings = HeimdallrSettings()
        for k, v in flask.request.form.items():
            if v != "":
                setattr(settings, k, v)
    else:
        print("Not in development mode")
    return flask.redirect("/")


def on_message(client: Any, userdata: Any, result: mqtt.client.MQTTMessage) -> None:
    """description"""
    global LOG_WRITER
    global GPU_STATS
    global KEEP_ALIVE
    d = json.loads(result.payload)
    keys = d.keys()
    for key in keys:
        if "gpu_stats" in d[key]:
            GPU_STATS[key] = d[key]["gpu_stats"]
            DU_STATS[key] = d[key]["du_stats"]
        else:
            GPU_STATS[key] = d[key]  # ["gpu_stats"]
            DU_STATS[key] = {}
        KEEP_ALIVE[key] = 0
    LOG_WRITER(
        f"received payload for {keys}, retain:{result.retain}, timestamp:{result.timestamp}"
    )


def on_disconnect(client: Any, userdata: Any, rc: Any) -> None:
    """description"""
    if rc != 0:
        print("Unexpected MQTT disconnection. Will auto-reconnect")
        client.reconnect()
        client.subscribe(ALL_CONSTANTS.MQTT_TOPIC, ALL_CONSTANTS.MQTT_QOS)


def setup_mqtt_connection(settings) -> None:
    """

    :param settings:
    :type settings:
    """
    if (
        False
        # settings.mqtt_access_token and False
    ):  # TODO: not implemented
        pass
        # MQTT_CLIENT.username_pw_set(settings.MQTT_ACCESS_TOKEN)
    else:
        MQTT_CLIENT.username_pw_set(
            settings.mqtt_username,
            settings.mqtt_password,
        )
    try:
        MQTT_CLIENT.connect(
            settings.mqtt_broker,
            settings.mqtt_port,
            keepalive=60,
        )
        MQTT_CLIENT.subscribe(ALL_CONSTANTS.MQTT_TOPIC, ALL_CONSTANTS.MQTT_QOS)
    except Exception as e:
        LOG_WRITER(f"MQTT connection error: {e}")
        # raise e


def main(
    *args,
    setting_scope: SettingScopeEnum = SettingScopeEnum.user,
    development=False,
    **kwargs,
) -> None:
    """description"""
    global LOG_WRITER, DEVELOPMENT

    if setting_scope == SettingScopeEnum.user:
        LOG_WRITER = LogWriter(
            ensure_existence(PROJECT_APP_PATH.user_log) / f"{PROJECT_NAME}_server.log"
        )
    else:
        LOG_WRITER = LogWriter(
            ensure_existence(PROJECT_APP_PATH.site_log) / f"{PROJECT_NAME}_server.log"
        )
    LOG_WRITER.open()
    MQTT_CLIENT.on_message = on_message
    MQTT_CLIENT.on_disconnect = on_disconnect

    if True:
        crystallised_heimdallr_settings = HeimdallrSettings(setting_scope)
        setup_mqtt_connection(settings=crystallised_heimdallr_settings)

    DASH_APP.title = ALL_CONSTANTS.HTML_TITLE
    DASH_APP.update_title = ALL_CONSTANTS.HTML_TITLE
    host = ALL_CONSTANTS.SERVER_ADDRESS
    port = ALL_CONSTANTS.SERVER_PORT

    DEVELOPMENT = development
    DASH_APP.layout = get_root_layout(DEVELOPMENT)

    if development:
        DASH_APP.run_server(
            host=host,
            port=port,
            debug=ALL_CONSTANTS.DEBUG,
            dev_tools_hot_reload=ALL_CONSTANTS.DEBUG,
        )
    else:
        # DASH_APP.run_server(host=host, port=port)
        serve(DASH_APP.server, **kwargs)

    LOG_WRITER.close()


if __name__ == "__main__":
    main(development=True)



import os
import sys
import time
from distutils.version import LooseVersion

import pytest

SKIP_TEST = True

if sys.platform == "linux" and not SKIP_TEST:
    print("Linux system detected")
    import pynvml

    NVML_PCIE_UTIL_TX_BYTES = pynvml.NVML_PCIE_UTIL_TX_BYTES
    NVML_PCIE_UTIL_RX_BYTES = pynvml.NVML_PCIE_UTIL_RX_BYTES
    NVML_PCIE_UTIL_COUNT = pynvml.NVML_PCIE_UTIL_COUNT

    # Test pynvml.nvmlDeviceGetNvLinkCapability

    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.mark.parametrize(
        "cap_type",
        [
            pynvml.NVML_NVLINK_CAP_P2P_SUPPORTED,  # P2P over NVLink is supported
            pynvml.NVML_NVLINK_CAP_SYSMEM_ACCESS,  # Access to system memory is supported
            pynvml.NVML_NVLINK_CAP_P2P_ATOMICS,  # P2P atomics are supported
            pynvml.NVML_NVLINK_CAP_SYSMEM_ATOMICS,  # System memory atomics are supported
            pynvml.NVML_NVLINK_CAP_SLI_BRIDGE,  # SLI is supported over this link
            pynvml.NVML_NVLINK_CAP_VALID,
        ],
    )  # Link is supported on this device
    def test_nvml_nvlink_capability(ngpus, handles, cap_type):
        for i in range(ngpus):
            for j in range(pynvml.NVML_NVLINK_MAX_LINKS):
                cap = pynvml.nvmlDeviceGetNvLinkCapability(handles[i], j, cap_type)
                assert cap >= 0

    # Test pynvml.nvmlDeviceResetNvLinkErrorCounters
    # Test pynvml.nvmlDeviceGetNvLinkErrorCounter
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.mark.parametrize(
        "error_type",
        [
            pynvml.NVML_NVLINK_ERROR_DL_REPLAY,
            pynvml.NVML_NVLINK_ERROR_DL_RECOVERY,
            pynvml.NVML_NVLINK_ERROR_DL_CRC_FLIT,
            pynvml.NVML_NVLINK_ERROR_DL_CRC_DATA,
        ],
    )
    def test_nvml_nvlink_error_counters(ngpus, handles, error_type):
        for i in range(ngpus):
            for j in range(pynvml.NVML_NVLINK_MAX_LINKS):
                assert (
                    pynvml.nvmlDeviceResetNvLinkErrorCounters(handles[i], j)
                    == pynvml.NVML_SUCCESS
                )
                error_count = pynvml.nvmlDeviceGetNvLinkErrorCounter(
                    handles[i], j, error_type
                )
                assert error_count >= 0

    # Fixture to initialize and finalize nvml
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.fixture(scope="module")
    def nvml(request):
        """description"""
        pynvml.nvmlInit()

        def nvml_close():
            """description"""
            pynvml.nvmlShutdown()

        request.addfinalizer(nvml_close)

    # Get GPU count
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.fixture
    def ngpus(nvml):
        """description"""
        result = pynvml.nvmlDeviceGetCount()
        assert result > 0
        print("[" + str(result) + " GPUs]", end=" ")
        return result

    # Get handles using pynvml.nvmlDeviceGetHandleByIndex
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.fixture
    def handles(ngpus):
        """description"""
        handles = [pynvml.nvmlDeviceGetHandleByIndex(i) for i in range(ngpus)]
        assert len(handles) == ngpus
        return handles

    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.fixture
    def serials(ngpus, handles):
        """description"""
        serials = [pynvml.nvmlDeviceGetSerial(handles[i]) for i in range(ngpus)]
        assert len(serials) == ngpus
        return serials

    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.fixture
    def uuids(ngpus, handles):
        """description"""
        uuids = [pynvml.nvmlDeviceGetUUID(handles[i]) for i in range(ngpus)]
        assert len(uuids) == ngpus
        return uuids

    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.fixture
    def pci_info(ngpus, handles):
        """description"""
        pci_info = [pynvml.nvmlDeviceGetPciInfo(handles[i]) for i in range(ngpus)]
        assert len(pci_info) == ngpus
        return pci_info

    ## ---------------------------- ##
    ## Start Fucntion-pointer Tests ##
    ## ---------------------------- ##

    # Test pynvml.nvmlSystemGetNVMLVersion
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvml_system_get_nvml_version(nvml):
        vsn = 0.0
        vsn = pynvml.nvmlSystemGetNVMLVersion().decode()
        print("[NVML Version: " + vsn + "]", end=" ")
        assert vsn > LooseVersion("0.0")

    # Test pynvml.nvmlSystemGetProcessName
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvml_system_get_process_name(nvml):
        procname = None
        procname = pynvml.nvmlSystemGetProcessName(os.getpid())
        print("[Process: " + str(procname.decode()) + "]", end=" ")
        assert procname != None

    # Test pynvml.nvmlSystemGetDriverVersion
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvml_system_get_driver_version(nvml):
        vsn = 0.0
        vsn = pynvml.nvmlSystemGetDriverVersion().decode()
        print("[Driver Version: " + vsn + "]", end=" ")
        assert vsn > LooseVersion("0.0")  # Developing with 396.44

    ## Unit "Get" Functions (Skipping for now) ##

    ## Device "Get" Functions (Note: Some functions are fixtures, above) ##

    # Test pynvml.nvmlDeviceGetHandleBySerial
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvml_device_get_handle_by_serial(ngpus, serials):
        handles = [pynvml.nvmlDeviceGetHandleBySerial(serials[i]) for i in range(ngpus)]
        assert len(handles) == ngpus

    # Test pynvml.nvmlDeviceGetHandleByUUID
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvml_device_get_handle_by_uuid(ngpus, uuids):
        handles = [pynvml.nvmlDeviceGetHandleByUUID(uuids[i]) for i in range(ngpus)]
        assert len(handles) == ngpus

    # Test pynvml.nvmlDeviceGetHandleByPciBusId
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvml_device_get_handle_by_pci_bus_id(ngpus, pci_info):
        handles = [
            pynvml.nvmlDeviceGetHandleByPciBusId(pci_info[i].busId)
            for i in range(ngpus)
        ]
        assert len(handles) == ngpus

    # [Skipping] pynvml.nvmlDeviceGetName
    # [Skipping] pynvml.nvmlDeviceGetBoardId
    # [Skipping] pynvml.nvmlDeviceGetMultiGpuBoard
    # [Skipping] pynvml.nvmlDeviceGetBrand
    # [Skipping] pynvml.nvmlDeviceGetCpuAffinity
    # [Skipping] pynvml.nvmlDeviceSetCpuAffinity
    # [Skipping] pynvml.nvmlDeviceClearCpuAffinity
    # [Skipping] pynvml.nvmlDeviceGetMinorNumber
    # [Skipping] pynvml.nvmlDeviceGetUUID
    # [Skipping] pynvml.nvmlDeviceGetInforomVersion
    # [Skipping] pynvml.nvmlDeviceGetInforomImageVersion
    # [Skipping] pynvml.nvmlDeviceGetInforomConfigurationChecksum
    # [Skipping] pynvml.nvmlDeviceValidateInforom
    # [Skipping] pynvml.nvmlDeviceGetDisplayMode
    # [Skipping] pynvml.nvmlDeviceGetPersistenceMode
    # [Skipping] pynvml.nvmlDeviceGetClockInfo
    # [Skipping] pynvml.nvmlDeviceGetMaxClockInfo
    # [Skipping] pynvml.nvmlDeviceGetApplicationsCloc
    # [Skipping] pynvml.nvmlDeviceGetDefaultApplicationsClock
    # [Skipping] pynvml.nvmlDeviceGetSupportedMemoryClocks
    # [Skipping] pynvml.nvmlDeviceGetSupportedGraphicsClocks
    # [Skipping] pynvml.nvmlDeviceGetFanSpeed
    # [Skipping] pynvml.nvmlDeviceGetTemperature
    # [Skipping] pynvml.nvmlDeviceGetTemperatureThreshold
    # [Skipping] pynvml.nvmlDeviceGetPowerState
    # [Skipping] pynvml.nvmlDeviceGetPerformanceState
    # [Skipping] pynvml.nvmlDeviceGetPowerManagementMode
    # [Skipping] pynvml.nvmlDeviceGetPowerManagementLimit
    # [Skipping] pynvml.nvmlDeviceGetPowerManagementLimitConstraints
    # [Skipping] pynvml.nvmlDeviceGetPowerManagementDefaultLimit
    # [Skipping] pynvml.nvmlDeviceGetEnforcedPowerLimit

    # Test pynvml.nvmlDeviceGetPowerUsage
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvmlDeviceGetPowerUsage(ngpus, handles):
        for i in range(ngpus):
            power_mWatts = pynvml.nvmlDeviceGetPowerUsage(handles[i])
            assert power_mWatts >= 0.0

    # Test pynvml.nvmlDeviceGetTotalEnergyConsumption
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvmlDeviceGetTotalEnergyConsumption(ngpus, handles):
        for i in range(ngpus):
            energy_mJoules1 = pynvml.nvmlDeviceGetTotalEnergyConsumption(handles[i])
            for j in range(10):  # idle for 150 ms
                time.sleep(0.015)  # and check for increase every 15 ms
                energy_mJoules2 = pynvml.nvmlDeviceGetTotalEnergyConsumption(handles[i])
                assert energy_mJoules2 >= energy_mJoules1
                if energy_mJoules2 > energy_mJoules1:
                    break
            else:
                assert False, "energy did not increase across 150 ms interval"

    # [Skipping] pynvml.nvmlDeviceGetGpuOperationMode
    # [Skipping] pynvml.nvmlDeviceGetCurrentGpuOperationMode
    # [Skipping] pynvml.nvmlDeviceGetPendingGpuOperationMode

    # Test pynvml.nvmlDeviceGetMemoryInfo
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvmlDeviceGetMemoryInfo(ngpus, handles):
        for i in range(ngpus):
            meminfo = pynvml.nvmlDeviceGetMemoryInfo(handles[i])
            assert (meminfo.used <= meminfo.total) and (meminfo.free <= meminfo.total)

    # [Skipping] pynvml.nvmlDeviceGetBAR1MemoryInfo
    # [Skipping] pynvml.nvmlDeviceGetComputeMode
    # [Skipping] pynvml.nvmlDeviceGetEccMode
    # [Skipping] pynvml.nvmlDeviceGetCurrentEccMode (Python API Addition)
    # [Skipping] pynvml.nvmlDeviceGetPendingEccMode (Python API Addition)
    # [Skipping] pynvml.nvmlDeviceGetTotalEccErrors
    # [Skipping] pynvml.nvmlDeviceGetDetailedEccErrors
    # [Skipping] pynvml.nvmlDeviceGetMemoryErrorCounter

    # Test pynvml.nvmlDeviceGetUtilizationRates
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvmlDeviceGetUtilizationRates(ngpus, handles):
        for i in range(ngpus):
            urate = pynvml.nvmlDeviceGetUtilizationRates(handles[i])
            assert urate.gpu >= 0
            assert urate.memory >= 0

    # [Skipping] pynvml.nvmlDeviceGetEncoderUtilization
    # [Skipping] pynvml.nvmlDeviceGetDecoderUtilization
    # [Skipping] pynvml.nvmlDeviceGetPcieReplayCounter
    # [Skipping] pynvml.nvmlDeviceGetDriverModel
    # [Skipping] pynvml.nvmlDeviceGetCurrentDriverModel
    # [Skipping] pynvml.nvmlDeviceGetPendingDriverModel
    # [Skipping] pynvml.nvmlDeviceGetVbiosVersion
    # [Skipping] pynvml.nvmlDeviceGetComputeRunningProcesses
    # [Skipping] pynvml.nvmlDeviceGetGraphicsRunningProcesses
    # [Skipping] pynvml.nvmlDeviceGetAutoBoostedClocksEnabled
    # [Skipping] nvmlUnitSetLedState
    # [Skipping] pynvml.nvmlDeviceSetPersistenceMode
    # [Skipping] pynvml.nvmlDeviceSetComputeMode
    # [Skipping] pynvml.nvmlDeviceSetEccMode
    # [Skipping] pynvml.nvmlDeviceClearEccErrorCounts
    # [Skipping] pynvml.nvmlDeviceSetDriverModel
    # [Skipping] pynvml.nvmlDeviceSetAutoBoostedClocksEnabled
    # [Skipping] pynvml.nvmlDeviceSetDefaultAutoBoostedClocksEnabled
    # [Skipping] pynvml.nvmlDeviceSetApplicationsClocks
    # [Skipping] pynvml.nvmlDeviceResetApplicationsClocks
    # [Skipping] pynvml.nvmlDeviceSetPowerManagementLimit
    # [Skipping] pynvml.nvmlDeviceSetGpuOperationMode
    # [Skipping] nvmlEventSetCreate
    # [Skipping] pynvml.nvmlDeviceRegisterEvents
    # [Skipping] pynvml.nvmlDeviceGetSupportedEventTypes
    # [Skipping] nvmlEventSetWait
    # [Skipping] nvmlEventSetFree
    # [Skipping] pynvml.nvmlDeviceOnSameBoard
    # [Skipping] pynvml.nvmlDeviceGetCurrPcieLinkGeneration
    # [Skipping] pynvml.nvmlDeviceGetMaxPcieLinkGeneration
    # [Skipping] pynvml.nvmlDeviceGetCurrPcieLinkWidth
    # [Skipping] pynvml.nvmlDeviceGetMaxPcieLinkWidth
    # [Skipping] pynvml.nvmlDeviceGetSupportedClocksThrottleReasons
    # [Skipping] pynvml.nvmlDeviceGetCurrentClocksThrottleReasons
    # [Skipping] pynvml.nvmlDeviceGetIndex
    # [Skipping] pynvml.nvmlDeviceGetAccountingMode
    # [Skipping] pynvml.nvmlDeviceSetAccountingMode
    # [Skipping] pynvml.nvmlDeviceClearAccountingPids
    # [Skipping] pynvml.nvmlDeviceGetAccountingStats
    # [Skipping] pynvml.nvmlDeviceGetAccountingPids
    # [Skipping] pynvml.nvmlDeviceGetAccountingBufferSize
    # [Skipping] pynvml.nvmlDeviceGetRetiredPages
    # [Skipping] pynvml.nvmlDeviceGetRetiredPagesPendingStatus
    # [Skipping] pynvml.nvmlDeviceGetAPIRestriction
    # [Skipping] pynvml.nvmlDeviceSetAPIRestriction
    # [Skipping] pynvml.nvmlDeviceGetBridgeChipInfo
    # [Skipping] pynvml.nvmlDeviceGetSamples
    # [Skipping] pynvml.nvmlDeviceGetViolationStatus

    # Test pynvml.nvmlDeviceGetPcieThroughput
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvmlDeviceGetPcieThroughput(ngpus, handles):
        for i in range(ngpus):
            tx_bytes_tp = pynvml.nvmlDeviceGetPcieThroughput(
                handles[i], NVML_PCIE_UTIL_TX_BYTES
            )
            assert tx_bytes_tp >= 0
            rx_bytes_tp = pynvml.nvmlDeviceGetPcieThroughput(
                handles[i], NVML_PCIE_UTIL_RX_BYTES
            )
            assert rx_bytes_tp >= 0
            count_tp = pynvml.nvmlDeviceGetPcieThroughput(
                handles[i], NVML_PCIE_UTIL_COUNT
            )
            assert count_tp >= 0

    # [Skipping] pynvml.nvmlSystemGetTopologyGpuSet
    # [Skipping] pynvml.nvmlDeviceGetTopologyNearestGpus
    # [Skipping] pynvml.nvmlDeviceGetTopologyCommonAncestor

    # Test pynvml.nvmlDeviceGetNvLinkVersion
    # Test pynvml.nvmlDeviceGetNvLinkState
    # Test pynvml.nvmlDeviceGetNvLinkRemotePciInfo
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    def test_nvml_nvlink_properties(ngpus, handles):
        for i in range(ngpus):
            for j in range(pynvml.NVML_NVLINK_MAX_LINKS):
                version = pynvml.nvmlDeviceGetNvLinkVersion(handles[i], j)
                assert version >= 1
                state = pynvml.nvmlDeviceGetNvLinkState(handles[i], j)
                assert state >= 0
                pci_info = pynvml.nvmlDeviceGetNvLinkRemotePciInfo(handles[i], j)
                assert isinstance(pci_info, pynvml.c_nvmlPciInfo_t)

    # Test pynvml.nvmlDeviceResetNvLinkUtilizationCounter
    # Test pynvml.nvmlDeviceSetNvLinkUtilizationControl
    # Test pynvml.nvmlDeviceGetNvLinkUtilizationCounter
    # Test pynvml.nvmlDeviceGetNvLinkUtilizationControl
    # Test pynvml.nvmlDeviceFreezeNvLinkUtilizationCounter
    @pytest.mark.skipif(sys.platform != "linux", reason="Test only on linux")
    @pytest.mark.parametrize("counter", [0, 1])
    @pytest.mark.parametrize("control", [0, 1, 2])
    def test_nvml_nvlink_counters(ngpus, handles, counter, control):
        reset = 0
        for i in range(ngpus):
            for j in range(pynvml.NVML_NVLINK_MAX_LINKS):
                assert (
                    pynvml.nvmlDeviceResetNvLinkUtilizationCounter(
                        handles[i], j, counter
                    )
                    == pynvml.NVML_SUCCESS
                )
                pynvml.nvmlDeviceSetNvLinkUtilizationControl(
                    handles[i], j, counter, control, reset
                )
                countdict = pynvml.nvmlDeviceGetNvLinkUtilizationCounter(
                    handles[i], j, counter
                )
                ctl = pynvml.nvmlDeviceGetNvLinkUtilizationControl(
                    handles[i], j, counter
                )
                assert countdict["rx"] >= 0
                assert countdict["tx"] >= 0
                assert ctl == control
                assert (
                    pynvml.nvmlDeviceFreezeNvLinkUtilizationCounter(
                        handles[i], j, counter, 1
                    )
                    == pynvml.NVML_SUCCESS
                )
                assert (
                    pynvml.nvmlDeviceFreezeNvLinkUtilizationCounter(
                        handles[i], j, counter, 0
                    )
                    == pynvml.NVML_SUCCESS
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian"
__doc__ = r"""

           Created on 29/03/2020
           """

import calendar

# https://github.com/pulls?q=is%3Aopen+is%3Apr+user%3Aaivclab+archived%3Afalse+
# https://github.com/pulls?user=aivclab
from datetime import datetime, timedelta
from enum import Enum
from typing import Generator, Iterable, List, Union

from github import Github  # pip install PyGithub
from github.GithubObject import _NotSetType
from github.Issue import Issue
from github.Label import Label
from github.Milestone import Milestone
from github.NamedUser import NamedUser
from github.PullRequest import PullRequest
from github.Repository import Repository
from sorcery import assigned_names


# TODO: CONVERT TO PLUGINS for heimdallr!


class RepoTypeEnum(Enum):
    all, public, private, forks, sources, member = assigned_names()
    internal = "internal"  # TODO: maybe not be supported


class RepoSortEnum(Enum):
    created, updated, pushed, full_name = assigned_names()


class DirectionEnum(Enum):
    asc, desc = assigned_names()


class StateEnum(Enum):
    open, closed, all = assigned_names()


class PullRequestSortEnum(Enum):
    (created, updated, popularity) = assigned_names()  # (comment  count),

    long_running = (
        "long-running"  # (age, filtering by pulls  updated in the  last  month).
    )


class IssueSortEnum(Enum):
    created, updated, comments = assigned_names()


def yield_org_repos(
    g: Github,
    org_name: str,
    repo_type: Union[str, RepoTypeEnum] = RepoTypeEnum.all,
    sort: Union[str, RepoSortEnum] = RepoSortEnum.pushed,
    direction: Union[str, DirectionEnum] = DirectionEnum.desc,
) -> Generator[Repository, None, None]:
    """

    Read more: https://docs.github.com/en/rest/reference/repos#list-organization-repositories--parameters

    Args:
      g:
      org_name:
      repo_type:
      sort:
      direction:

    Returns:

    """
    repo_type = RepoTypeEnum(repo_type)
    sort = RepoSortEnum(sort)
    direction = DirectionEnum(direction)
    yield from g.get_organization(org_name).get_repos(
        type=repo_type.value, sort=sort.value, direction=direction.value
    )


def yield_org_issues(g: Github, org_name: str) -> Generator[Issue, None, None]:
    """

    :param g:
    :param org_name:
    """
    yield from yield_issues(yield_org_repos(g, org_name))


def yield_org_prs(g: Github, org_name: str) -> Generator[PullRequest, None, None]:
    """

    :param g:
    :param org_name:
    """
    yield from yield_prs(yield_org_repos(g, org_name))


def yield_issues(
    iterable: Iterable[Repository],
    milestone: Union[Milestone, str] = _NotSetType,
    state: Union[str, StateEnum] = StateEnum.open,
    assignee: Union[NamedUser, str] = _NotSetType,
    mentioned: NamedUser = _NotSetType,
    labels: Union[List[str], List[Label]] = _NotSetType,
    sort: Union[str, IssueSortEnum] = IssueSortEnum.updated,
    direction: Union[str, DirectionEnum] = DirectionEnum.desc,
    since: datetime = _NotSetType,  # datetime.now(),
    creator: Union[str, NamedUser] = _NotSetType,
) -> Generator[Issue, None, None]:
    """

    Read more: https://docs.github.com/en/rest/reference/issues#list-issues-assigned-to-the-authenticated-user--parameters

    Args:
      iterable:
      milestone:
      state:
      assignee:
      mentioned:
      labels:
      sort:
      direction:
      since:
      creator:

    Returns:

    """
    state = StateEnum(state)
    sort = IssueSortEnum(sort)
    direction = DirectionEnum(direction)
    for repo in iterable:
        yield from repo.get_issues(
            milestone=milestone,
            state=state.value,
            assignee=assignee,
            mentioned=mentioned,
            labels=labels,
            sort=sort.value,
            direction=direction.value,
            since=since,
            creator=creator,
        )


def yield_prs(
    iterable: Iterable[Repository],
    state: Union[str, StateEnum] = StateEnum.open,
    sort: Union[str, PullRequestSortEnum] = PullRequestSortEnum.updated,
    direction: Union[str, DirectionEnum] = DirectionEnum.desc,
    base: str = _NotSetType,
    head: str = _NotSetType,
) -> Generator[PullRequest, None, None]:
    """

    Read more: https://docs.github.com/en/rest/reference/pulls#list-pull-requests--parameters

    Args:
      iterable:
      state:
      sort:
      direction:
      base:
      head:

    Returns:

    """
    state = StateEnum(state)
    sort = PullRequestSortEnum(sort)
    direction = DirectionEnum(direction)
    for repo in iterable:
        yield from repo.get_pulls(
            state=state.value,
            sort=sort.value,
            direction=direction.value,
            base=base,
            head=head,
        )


#
def get_pull_request(g, org_name, repo_name, pull_request_number):
    """

    :param g:
    :param org_name:
    :param repo_name:
    :param pull_request_number:
    :return:
    """
    return (
        g.get_organization(org_name).get_repo(repo_name).get_pull(pull_request_number)
    )


def utc_to_epoch(timestamp):  # Timestamp is a datetime object in UTC time
    """

    :param timestamp:
    :return:
    """
    return calendar.timegm(timestamp.utctimetuple())


if __name__ == "__main__":
    from heimdallr.configuration.heimdallr_settings import HeimdallrSettings

    gthb = Github(HeimdallrSettings().github_token)
    remaining = gthb.rate_limiting[0]
    reset_seconds = gthb.rate_limiting_resettime - utc_to_epoch(datetime.utcnow())
    interval_seconds = ((reset_seconds + 1) / (remaining + 1)) + 1
    print(f"intervals recommended {interval_seconds} seconds")
    if remaining > 0 and False:
        # g = Github(base_url="https://{hostname}/api/v3", login_or_token="access_token") # Github Enterprise with custom hostname
        # print(list(yield_org_repos(gthb, 'aivclab')))
        print(list(yield_org_issues(gthb, "aivclab"))[0].repository.name)
    else:
        print("rate limited")

        print(f"reset in {timedelta(seconds=reset_seconds)} ")
    # print(list(yield_org_prs(gthb, 'aivclab')))
